#include <glib.h>

#include "tuxmodule.h"

static GHashTable *queues;
static int queues_sem;

static GTrashStack *olditems;

typedef struct queue_item queue_item;
struct queue_item {
    queue_item *next;
    int sock;
};

typedef struct queue {
    queue_item *head;
    queue_item *tail;
    char *name;
    int sem;
} queue;



/* dequeue_first must be called with q->sem held
 * returns 0 if the queue was already empty
 * returns the socket dequeued otherwise
 */
static int
dequeue_first(queue *q) {
    queue_item *i = q->head;
    int sock;

    if (!q->head) return 0;

    sock = i->sock;
    q->head = i->next;
    g_trash_stack_push(&olditems, (gpointer) i);
    tux(TUX_ACTION_CONTINUE_REQ, (user_req_t *)sock);
    return sock;
}

static queue *
queue_from_name(char *queue_name) {
    queue *q;

    if (!queue_name) return (queue *)-1;

    TUXAPI_down(&queues_sem);
    q = g_hash_table_lookup(queues, queue_name);
    TUXAPI_up(&queues_sem);

    return q;
}




void *
TUXAPI_create_queue_anonymous(void) {
    queue *q = NULL;

    q = calloc(sizeof(queue), 1);

    return q;
}


void *
TUXAPI_create_queue(char *queue_name) {
    queue *q = NULL;

    if (!queue_name) return NULL;

    TUXAPI_down(&queues_sem);
    if (!queues)
	queues = g_hash_table_new(g_str_hash, g_str_equal);
    else
	q = g_hash_table_lookup(queues, queue_name);
    TUXAPI_up(&queues_sem);

    if (q) {
	return q;
    }

    q = TUXAPI_create_queue_anonymous();
    q->name = strdup(queue_name);

    TUXAPI_down(&queues_sem);
    g_hash_table_insert(queues, q->name, q);
    TUXAPI_up(&queues_sem);

    return q;
}

int
TUXAPI_remove_queue(void *queue_handle) {
    queue *q = queue_handle;

    if (!q) return -1;

    TUXAPI_down(&q->sem);
    while (dequeue_first(q));
    /* Leave the semaphore held -- we don't want a programmer error
     * to cause events to be lost.
     */

    if (q->name) {
	TUXAPI_down(&queues_sem);
	g_hash_table_remove(queues, q->name);
	TUXAPI_up(&queues_sem);
	free(q->name);
    }

    free(q);
    return 0;
}

int
TUXAPI_enqueue_request(user_req_t *req, void *queue_handle) {
    queue *q = queue_handle;
    queue_item *i;
    int ret;

    if (!req) return -1;
    i = g_trash_stack_pop(&olditems);
    if (!i)
	i = malloc(sizeof(queue_item));
    if (!i)
	return -1;
    i->sock = req->sock;
    i->next = NULL;

    ret = tux(TUX_ACTION_POSTPONE_REQ, req);

    TUXAPI_down(&q->sem);
    if (q->head) { /* something already on queue */
	q->tail->next = i;
	q->tail = i;
    } else { /* nothing on queue */
	q->head = q->tail = i;
    }
    TUXAPI_up(&q->sem);

    return ret;
}

int
TUXAPI_enqueue_request_by_name(user_req_t *req, char *queue_name) {
    queue *q = queue_from_name(queue_name);

    if (q < 0 || !q) return (int)q;

    TUXAPI_enqueue_request(req, q);
}

int
TUXAPI_dequeue_request(void *queue_handle) {
    queue *q = queue_handle;
    int sock;

    if (!q) return -1;

    TUXAPI_down(&q->sem);
    sock = dequeue_first(q);
    TUXAPI_up(&q->sem);

    return sock;
}

int
TUXAPI_dequeue_request_by_name(char *queue_name) {
    queue *q = queue_from_name(queue_name);

    if (q < 0 || !q) return (int)q;

    return TUXAPI_dequeue_request(q);
}
